/** @file
@code
***************************************************************************
    CVS                  : $RCSfile: demo_db.cpp,v $
                           $Revision: 1.1 $

***************************************************************************
@endcode*/

#include "demo_db.h"

//======MAIN========================================================================
/**
 * Main program function.
 * Uses the demo_db object for the actual program execution.
 * When the help or version options are used, the app exits directly after showing the info.
 * @return -1 on error.
 */
int main(int argc, char* argv[]) {
	try {
		bool Stop=false;
		demo_db DemoApp;
		// Init ; give command-line options to demo_db object to handle
		Stop = DemoApp.Init(argc, argv);
		if (!Stop) {
			cout << VERSIONTEXT;
			// Open database-text-file and load into memory
			DemoApp.OpenDatabase();
			// open formatting files for DS175 and PC2
			DemoApp.OpenFormatFiles();
			// Create and Connect socket
			DemoApp.StartSocket();
			// Start execution
			DemoApp.Execution();
		};
		return 0;
	} catch(string Error) {
		if (!Error.empty()) {
			cerr << "ERROR: " << Error << endl;
		} else {
			cerr << "ERROR: Unkown error..." << endl;
			throw;
		};
		return -1;
	};

};

//======DEMO_DB object=============================================================
/**
 * Constructor. Initializes strings, ints and bools; only to make sure they have
 * an "empty" value.
 * @return
 */
demo_db::demo_db() {
	//Initialize everything, just to make sure...
	DBFilename = "";
	ServerPort=0;
	ServerIP="";
	InputBuffer="";
	Exit=false;
	PC2Format="";
	DS175Format="";
	command="";
	prompt=false;
}

/**
 * Destructor is empty
 * @return
 */
demo_db::~demo_db() {}


/**
 * Set some defaults and process any commandline options.
 * The defaults for the databasefilename, the server-IP and server-PORT are defined in the
 * demo_header.h file.
 * The -h and -v options cause this function to return true; which can be used to terminate the app.
 * @param argc
 * @param argv[]
 * @return bool (Exit)
 */
bool demo_db::Init(int argc, char* argv[]) {
	string option, optname, optdata;
	int pos;
	Exit = false;
	//defaults
	DBFilename = def_DBFILE;
	ServerPort = def_PORT;
	ServerIP = def_IP;
	//read command-line options and process them
	for (int i=1; i<argc ; ++i) {
		option = argv[i];
		if ( (pos = option.find('=')) >1 ) {
			optname = option.substr(0,pos);
			optdata = option.substr(pos+1);
		} else {
			optname=option;
		};
		if (optname[0] == '-') {
			switch (optname[1]) {
				case 'h':
					cout << VERSIONTEXT << HELPTEXT ;
					Exit = true;
					break;
				case 'v':
					cout << VERSIONTEXT ;
					Exit = true;
					break;
				case 'f':
					if (!optdata.empty()) {
						DBFilename = optdata;
					};
					break;
				case 'i':
					if (!optdata.empty()) {
						ServerIP = optdata;
					};
					break;
				case 'p':
					if (!optdata.empty()) {
						ServerPort = atoi(optdata.c_str());
					};
					break;
			}
		};
	};
	return Exit;
}

/**
 * Opens the database, using its Load() function
 * throws an error when DBFilename is empty (which should at least have a default value).
 */
void demo_db::OpenDatabase(string name) {
	if (name.empty() && DBFilename.empty()) {
		throw "OpenDatabase: Empty Database filename.";
	};
	if (!name.empty()) {
		DBFilename = name;
	};
	DB.Load(DBFilename);
	cout<<"Database: " << DBFilename << " opened." << endl;
}

/**
 * Load the formatting files for the DS175 and PC2.
 * The filenames are defined in the demo_header.h file (DS175FILENAME & PC2FILENAME)
 * Sends a warning to stderr when a file is not found, but continues execution.
 * (results in unformatted data being send to PCxx's)
 * These files use the string \<ESC\> to represent the escape character for the PCxx-devices (\\x1b).
 */
void demo_db::OpenFormatFiles() {
	char c;
	//DS175------------------------------
	ifstream DS175file(DS175FILENAME);
	if (DS175file) {
		DS175Format="";
		while ( (c=DS175file.get())!= EOF) {
			DS175Format += c;
		}
// 		DS175Format = ExpandEsc(DS175Format);
		DS175Format = ExpandAll(DS175Format);
		cout << ERRORTEXT01 << endl;
		DS175file.close();
	} else {
		cerr << ERRORTEXT02 << DS175FILENAME <<") load: Unable to open file.\n" << endl;
	};

	//PC2------------------------------
	ifstream PC2file(PC2FILENAME);
	if (PC2file) {
		PC2Format="";
		while ( (c=PC2file.get())!= EOF) {
			PC2Format += c;
		}
// 		PC2Format = ExpandEsc(PC2Format);
		PC2Format = ExpandAll(PC2Format);
		cout << ERRORTEXT03 << endl;
		PC2file.close();
	} else {
		cerr << ERRORTEXT04 << PC2FILENAME <<") load: Unable to open file.\n" << endl;
	};
}

/**
 * Connects to an PC_Ethernet_Driver using the ServerIP and ServerPort values,
 * which default to the standard port on localhost.
 */
void demo_db::StartSocket() {
	DBClientSocket.Connect(ServerIP, ServerPort);
	cout << INFOTEXT01 << endl;
}

/**
 * Main loop.
 * Process input from the PC_Ethernet_Driver, and manage the commandprompt.
 * Extracts Tags from input with ExtractTag() and processes them with ProcessRequest().
 */
void demo_db::Execution() {
	string Answer="";
	command="";
	int i=0;
	prompt=true;
	bool Quit = false;
	struct TagType Tag;
	while (!Quit) {
		//Process input
		InputBuffer += DBClientSocket.Read();
		if (!InputBuffer.empty()) {
			Tag = ExtractTag(InputBuffer);
			if ( Tag.Name != "" ) {
				Answer = ProcessRequest(Tag);
				if (!Answer.empty()) {
					DBClientSocket.Write(Answer);
				};
			};
		}else{
		//Suspend for a short while
			mSleep(1);
		};
		//Process Command line
		Quit = CommandParser();
	};
}

/**
 * Process a Tag according to its name (currently only request).
 * The database is searched for the data from the tag (usually a barcode)
 * The answer from the database is then inserted into the format-string, depending
 * on the type found in the tag-options, using the FormatDS175() and FormatPC2() functions.
 * The resulting formatted answer is then packaged into a tag(-string) and returned to the calling function.
 * @param tag TagType struct, containing the name, options and data of a tag.
 * @return
 */
string demo_db::ProcessRequest(struct TagType tag) {
	string db_answer="", id, answer="", type, answer_text;
	if (tag.Name == "request") {
		db_answer = Database_Lookup(tag.Data);
		if (db_answer == "") {
			db_answer = Database_Lookup("unknown");
			if (db_answer == "") {
				return "";
			};
		};
		id = GetOption(tag.Options, "id");
		type = GetOption(tag.Options,"type");
		if (type==PC_DEVICETYPE1 || type==PC_DEVICETYPE1l) {
			answer_text = FormatDS175(db_answer);
		} else if (type==PC_DEVICETYPE2 || type==PC_DEVICETYPE2l) {
			answer_text = FormatPC2(db_answer);
		} else {
			answer_text = db_answer;
		};
		answer += "<data";
		if (id !="") {
			answer+=" id=" + id;
		};
		answer+=">" + answer_text + "</data>";
	};
	return answer;
}

/**
 * If a complete tag (starttag, data and endtag) is found in input, the tag is returned and removed from input.
 * Data outside a tag is discarded.
 * @param input
 * @return The first complete tag.
 */
struct TagType demo_db::ExtractTag(string& input) {
	struct TagType tag;
	tag.Name="";
	tag.Options="";
	tag.Data="";
	bool NoTag, endtagfound=false;
	string opentag, closetag;
	unsigned int pos, start, end, datastart, dataend;
	//look for starttag
	start=input.find('<');
	if (start != string::npos) {
		// if first tag is an end-tag (which is an error), delete it and leave.
		// if it really is an end-tag, part of it remains in the inputbuffer, but this is discarded on the next pass...
		if (input[start+1] == '/') {
			input = input.substr(start+2);
			return tag;
		};
		end = input.find('>', start);
		if (end != string::npos) {
			// start-tag found.
			opentag = input.substr(start+1, end-start-1);
			if ( (pos=opentag.find(' ')) != string::npos) {
				tag.Name = opentag.substr(0, pos);
				tag.Options = opentag.substr(pos+1);
			} else {
				tag.Name = opentag;
			};
			datastart = end+1;
			dataend = datastart;
			NoTag = false;
			//look for endtag
			while (!endtagfound && !NoTag) {
				if ( (start=input.find("</",dataend)) != string::npos) {
					if ( (end = input.find('>', start)) != string::npos) {
						// endtag found
						closetag = input.substr(start+1, end-start-1);
						if (closetag.empty() || ((closetag.find(tag.Name) != string::npos)) ) {
							// found matching (or generic) end-tag
							tag.Data = input.substr(datastart, start-datastart);
							input.erase(0,end+1);
							endtagfound = true;
						} else {
							//skip incorrect end-tag
							endtagfound = false;
							dataend = end;
						};
					}else{ //no correct end-tag exists.
						NoTag = true;
						tag.Name="";
						tag.Options="";
						tag.Data="";
					};
				} else { //no correct end-tag exists.
					NoTag = true;
					tag.Name="";
					tag.Options="";
					tag.Data="";
				};
			};
		};
	};
	return tag;
}

/**
 * Get the value of an option-item from the options-string.
 * The name and value of an item are separated by a '=' character.
 * @param options string containing one or more options separated by spaces.
 * @param item string: name of the item to be found in options.
 * @return string: value of the item; empty string if not found.
 */
string demo_db::GetOption(string options, string item) {
	string value, s;
	unsigned int start, end;
	s = item + '=';
	//find item
	start = options.find(s)+s.size();
	if ( start != string::npos) {
		//look for seperator
		if ((end=options.find(' ', start)) != string::npos) {
			value = options.substr(start, start-end-1);
		} else {
			value = options.substr(start);
		};
	} else {
		value = "";
	};
	return value;
}

/**
 * Lookup an item in the database and return its value.
 * @param key string to search for in database.
 * @return reply from the database.
 */
string demo_db::Database_Lookup(string key) {
	string answer="";
	answer = DB.Find(key);
	return answer;
}

/**
 * Replace names in the format string, indicated by {name}, with the corresponding data from the
 * database.
 * data contains lines of the format "name: value".
 * The result is obtained by replacing "{name}" with "value" in the Format-string.
 * If the Formatting string is empty, the data is returned without any formatting
 * @param data data form the database.
 * @return formatted string containing the items from data.
 */
string demo_db::FormatDS175(string data) {
	unsigned int i,j;
	string line, result,item, value, s;
	istringstream datastream(data);
	result = DS175Format;
	if (result.empty()) {
		return data;
	};
	//parse data line-by-line
	while ( getline(datastream, line) ) {
		//seperate item-name from its data
		i= line.find(':');
		if (i!= string::npos) {
			item = line.substr(0, i);
			value = line.substr(i+1);
			//fill in item into the DS175-format-string
			//an item may exist more than once in the format file...
			s = '{'+item+'}';
			j = result.find(s);
			while (j != string::npos) {
				result.replace(j, s.size(), value);
				j = result.find(s);
			};
		};
	};
	return result;
}

/**
 * Replace names in the format string, indicated by {name}, with the corresponding data from the
 * database.
 * data contains lines of the format "name: value".
 * The result is obtained by replacing "{name}" with "value" in the Format-string.
 * If the Formatting string is empty, the data is returned without any formatting
 * @param data data form the database.
 * @return formatted string containing the items from data.
 */
string demo_db::FormatPC2(string data) {
	unsigned int i,j;
	string line, result,item, value, s;
	istringstream datastream(data);
	result = PC2Format;
	if (result.empty()) {
		return data;
	};
	//parse data line-by-line
	while ( getline(datastream, line) ) {
		//seperate item-name from its data
		i= line.find(':');
		if (i!= string::npos) {
			item = line.substr(0, i);
			value = line.substr(i+1);
			//fill in item into the PC2-format-string
			s = '{'+item+'}';
			j = result.find(s);
			if (j != string::npos) {
				result.replace(j, s.size(), value);
			};
		};
	};
	return result;
}

/**
 * Display a prompt (when neccesary) and handle commands given by the user.
 * Commands:
 * - help: display a helptext, defined in demo_header.h
 * 		@see this helptext (CMDHELPTEXT) for more info on the available commands.
 * @return true when the quit command was given.
 */
bool demo_db::CommandParser() {
	int j;
	bool CmdOk;
	if (prompt) {
		write(1, "demo> ",6); //write to stdout (filedescriptor 1)
		prompt=false;
	};
	CmdOk = ReadCommand();
	if (CmdOk) {
		// QUIT
		if (command.find("quit") == 0) {
			return true;
			// HELP
		} else if ( (command.find("help")==0) || (command.find("?")==0) ) {
			cout << VERSIONTEXT << HELPCMDTEXT;
			// SEND
		} else if ( command.find("send ")==0 ) {
			j = command.find('\n');
			string filename = command.substr(5,j-5);
			string data="";
			ifstream file(filename.c_str());
			if (!file) {
				cerr << "Error in Send: File not found ("<< filename <<")\n";
				prompt=true;
                command="";
				return false;
			};
			char c;
			while ((c=file.get())!= EOF) {
				data += c;
			};
			file.close();
// 			data = ExpandEsc(data);
			data = ExpandAll(data);
			DBClientSocket.Write(data);
		} else if ( command.find("loadformat")==0 ) {
			OpenFormatFiles();
		} else if ( command.find("loaddb")==0 ) {
			try {
				int i;
				string name="";
				if (command.size()>7 && command[6]==' ') {
					i=command.find_first_of("\n\r\t ",7);
					name = command.substr(7,i-7);
				};
				OpenDatabase(name);
			} catch(string Error) {
				//if load fails issue warning but continue...
				cerr << "Warning: " << Error;
			};
		};
		command = "";
		prompt=true;
	};
	return false;
}

/**
 * NON-Blocking read from stdin
 * OR: read from keyboard (win32)
 * @return
 */
bool demo_db::ReadCommand() {
#ifdef WIN32
	char c;
	while (_kbhit()) {
		c = _getche();
		//the _getche() function returns a CR at the end of a line instead of a LF.
		// We replace it because in all our code a \n (LF) is used as the end-of-line.
		if (c=='\r') {
			c = '\n';
			//also echo the "\n"
			_putch(c);
		}
		command+=c;
	};
#else

	int DataReady,count;
	char buf[100];
	fd_set ReadSet;
	struct timeval timeout;
	// init the descriptor set
	FD_ZERO(&ReadSet);
	FD_SET(0, &ReadSet); // create descriptor set for select, descr 0 is stdin
	timeout.tv_sec = 0;
	timeout.tv_usec = 1;
	// wait for data to arrive
	DataReady = select(1, &ReadSet, NULL, NULL, &timeout);
	if (DataReady > 0) {
		// read data
		count = read(0, buf, 100 );
		if (count>0) {
			buf[count]=0;
			command += buf;
		};
	};
#endif

	if (command.find('\n') != string::npos) {
		return true;
	} else {
		return false;
	};
};
